﻿using Eingabeverwaltung.Metamodell;
using System.IO;
using System;
using Eingabeverwaltung.Parser;
using System.Collections.Immutable;
using System.Collections.Generic;

namespace Eingabeverwaltung
{
    public class Eingabeverwalter : AbstactEingabeverwalter
    {
        public Eingabeverwalter() : base() { }

        /// <summary>
        /// This Method manages the input without pre-inputed paths.
        /// </summary>
        public override void verwalteEingabe()
        {
            if (inputContainer == null) { initialEingabe(); return; }
            eingabe();
        }

        /// <summary>
        /// This Method manages the input, with pre-inputed paths.
        /// </summary>
        /// <param name="ADTPath"></param>
        /// <param name="TaskPath"></param>
        /// <param name="ProofPath"></param>
        public void verwalteEingabe(string ADTPath, string TaskPath, string ProofPath)
        {
            try
            {
                ADT_Eingabe(ADTPath);
                Task_Eingabe(TaskPath);
                Proof_Eingabe(ProofPath);
            }
            catch (FileNotFoundException f)
            {
                Console.WriteLine(f.Message);
                verwalteEingabe();
            }
            catch (DirectoryNotFoundException d)
            {
                Console.WriteLine(d.Message);
                verwalteEingabe();
            }
            catch (ArgumentException a)
            {
                Console.WriteLine(a.Message);
                verwalteEingabe();
            }
        }

        /// <summary>
        /// This Method manages the input if it is not the initial-input.
        /// </summary>
        private void eingabe()
        {
            ImmutableDictionary<string, Action> eingabe = new Dictionary<string, Action>
            {
                ["ADT"] = New_ADT_Eingabe,
                ["Task"] = New_Task_Eingabe,
                ["Proof"] = Proof_Eingabe,
                ["exit"] = Exit_Eingabe,
                ["printContainer"] = printContainer,
                ["showContainer"] = showContainer
            }.ToImmutableDictionary();
            string ausgabe = "Was möchten Sie eingeben? ";
            foreach (string input in eingabe.Keys)
                ausgabe += " " + input + " /";
            ausgabe = ausgabe.Remove(ausgabe.Length - 2);
            Console.WriteLine(ausgabe);
            string eingabeArt = Console.ReadLine();
            try {
                eingabe[eingabeArt](); 
            }
            catch (KeyNotFoundException e) { Console.WriteLine(e.Message); this.eingabe(); }
            catch (FileNotFoundException e)
            {
                Console.WriteLine(e.Message);
                verwalteEingabe();
            }
            catch (DirectoryNotFoundException e)
            {
                Console.WriteLine(e.Message);
                verwalteEingabe();
            }
            void New_ADT_Eingabe() => renewEingabe("");
            void New_Task_Eingabe() => renewEingabe("Task");
        }


        /// <summary>
        /// This Method manages the initial-input.
        /// </summary>
        /// 
        private void initialEingabe() => renewEingabe("");

        /// <summary>
        /// This Method renews the input. startAt can be "" to renew all input, "Task" to renew Task and Proof or "Proof" to renew only Proof.
        /// </summary>
        /// <param name="startAt"></param>
        private void renewEingabe(string startAt)
        {
            try
            {
                switch (startAt)
                {
                    case "Task":
                        Task_Eingabe();
                        Proof_Eingabe();
                        break;
                    case "Proof":
                        Proof_Eingabe();
                        break;
                    default:
                        ADT_Eingabe();
                        Task_Eingabe();
                        Proof_Eingabe();
                        break;
                }
            }
            catch (FileNotFoundException f)
            {
                Console.WriteLine(f.Message);
                verwalteEingabe();
            }
            catch (DirectoryNotFoundException d)
            {
                Console.WriteLine(d.Message);
                verwalteEingabe();
            }
            catch (ArgumentException a)
            {
                Console.WriteLine(a.Message);
                verwalteEingabe();
            }
        }
        public void Eingabe_Interface(string action, string path)
        {
            ImmutableDictionary<string, Action<string>> eingabe = new Dictionary<string, Action<string>>
            {
                ["ADT"] = ADT_Eingabe,
                ["Task"] = Task_Eingabe,
                ["Proof"] = Proof_Eingabe,
            }.ToImmutableDictionary();
            try
            {
                eingabe[action](path);
            }
            catch (FileNotFoundException)
            {
                Console.WriteLine("File not found on path \"" + path + "\".");
            }
            catch (DirectoryNotFoundException)
            {
                Console.WriteLine("Directory not found on path \"" + path + "\".");
            }
        }

        private void ADT_Eingabe() => ADT_Eingabe("");
        private void ADT_Eingabe(string path)
        {
            if (path == "")
            {
                Console.Write("Pfad zum ADT: ");
                path = Console.ReadLine();
            }
            string input = File.ReadAllText(path);
            try
            {
                ADT adt = ADT_Antlr.parse(input);
                inputContainer = new Container();
                inputContainer.ADT = adt;
            }
            catch (Exception e)
            {
                Console.WriteLine("Failed to parse the ADT at path \"" + path + "\"!");
                Console.WriteLine(e.Message + "\n" + e.StackTrace);
                verwalteEingabe();
            }
        }

        private void Task_Eingabe() => Task_Eingabe("");
        private void Task_Eingabe(string path) 
        {
            if (path == "")
            {
                Console.Write("Pfad zur Aufgabenstellung: ");
                path = Console.ReadLine();
            }
            string input = File.ReadAllText(path);
            try
            {
                Task_Antlr.parse(input,inputContainer);
                inputContainer.Proof = null;
            }
            catch (Exception e)
            {
                Console.WriteLine("Failed to parse the Task at path \"" + path + "\"!");
                Console.WriteLine(e.Message + "\n" + e.StackTrace);
                verwalteEingabe();
            }
        }

        private void Proof_Eingabe() => Proof_Eingabe("");
        private void Proof_Eingabe(string path) 
        {
            if (path == "")
            {
                Console.Write("Pfad zum Beweis: ");
                path = Console.ReadLine();
            }
            string input = File.ReadAllText(path);
            try
            {
                Proof_Antlr.parse(input, inputContainer);
            }
            catch (Exception e)
            {
                Console.WriteLine("Failed to parse the Proof at path \"" + path + "\"!");
                Console.WriteLine(e.Message + "\n" + e.StackTrace);
                verwalteEingabe();
            }
        }
        /// <summary>
        /// Does nothing in order to let the Eingabe exit.
        /// </summary>
        private void Exit_Eingabe() { }

        /// <summary>
        /// Prints Container with its metadata
        /// </summary>
        private void printContainer()
        {
            string print = "";
            try { print = inputContainer.ToString(true); }
            catch (NullReferenceException) { }
            Console.WriteLine(print);
        }

        /// <summary>
        /// Prints Container without metadata
        /// </summary>
        private void showContainer()
        {
            string print = "";
            try { print = inputContainer.ToString(false); }
            catch (NullReferenceException) { }
            Console.WriteLine(print);
        }
    }
}