﻿using Eingabeverwaltung.Metamodell;
using Eingabeverwaltung.Parser.ADT_Parser;
using System;
using Antlr4.Runtime.Tree;
using Antlr4.Runtime;
using Antlr4.Runtime.Misc;
using System.Collections.Generic;
using System.Linq;

namespace Eingabeverwaltung.Parser
{
    public class ADT_Listener : IADTListener
    {
        // object on which will be build
        private Container container;

        // helping objects for building
        private List<Sort> op_sorts;

        private Listener_Helper LH;

        public ADT_Listener(Container container)
        {
            this.container = container;
            op_sorts = null;

            LH = new Listener_Helper(null, null, null, null, null, null);
        }

        public void EnterAdt([NotNull] ADTParser.AdtContext context) { }
        public void ExitAdt([NotNull] ADTParser.AdtContext context)
        {
            if (LH.adt_sorts == null) throw new Exception("adt_sorts = null at ExitAdt");
            if (LH.adt_ops == null) throw new Exception("adt_ops = null at ExitAdt");
            if (LH.adt_axioms == null) LH.adt_axioms = new List<Axiom>();
            if (LH.adt_sorts.Count < 1) throw new Exception("adt_sorts is empty at ExitAdt");
            if (LH.adt_ops.Count < 1) throw new Exception("adt_ops is empty at ExitAdt");
            container.ADT = new ADT(LH.adt_sorts, LH.adt_ops, LH.adt_axioms);
        }
        public void EnterName([NotNull] ADTParser.NameContext context) { }
        public void ExitName([NotNull] ADTParser.NameContext context) { }

        public void EnterSorts([NotNull] ADTParser.SortsContext context) => LH.adt_sorts = new List<Sort>();
        public void ExitSorts([NotNull] ADTParser.SortsContext context)
        {
            if (LH.adt_sorts.Count < 1) throw new Exception("adt_sorts is empty at ExitSorts");
        }

        public void EnterSort([NotNull] ADTParser.SortContext context) { }
        public void ExitSort([NotNull] ADTParser.SortContext context)
        {
            string sortName = context.BEZ().GetText();
            if (context.Parent.GetChild(0).GetText() == "sorts")
            {
                Test_Sorts_all_unique(sortName);
                LH.adt_sorts.Add(new Sort(sortName));
                return;
            }
            if (new List<string>(){"operations","constructors"}.Contains(context.Parent.Parent.GetChild(0).GetText()))
            {
                Sort sort = LH.findSortInADT_Sorts(sortName);
                Test_Operation_all_Sorts_were_defined_in_ADT(sortName);
                op_sorts.Add(sort);
                return;
            }
            if (context.Parent.Parent.GetChild(0).GetText() == "vars") return; // this job is done in ExitVar
            throw new Exception("ExitSort with Context = " + context.GetText() + " in parent "+ context.Parent.GetChild(0).GetText() + " was not able to find the sense of its life.");
        }
        public void EnterConstructors([NotNull] ADTParser.ConstructorsContext context) => LH.adt_ops = new List<Operation>();
        public void ExitConstructors([NotNull] ADTParser.ConstructorsContext context)
        {
            if (LH.adt_ops.Count < 1) throw new Exception("adt_ops is empty at ExitConstructors");
            // check for each sort: numer & type of constructors is ok
            foreach (Sort sort in LH.adt_sorts)
                if (sort.GetBuildingConstructors().Count > 0 && sort.GetBasicConstructors().Count == 0)
                    throw new Exception("The sort "+sort.name+" has building constructors, but no basic constructors.");
        }

        public void EnterOperations([NotNull] ADTParser.OperationsContext context) 
        {
            if (LH.adt_ops == null) LH.adt_ops = new List<Operation>(); 
        }
        public void ExitOperations([NotNull] ADTParser.OperationsContext context)
        {
            if (LH.adt_ops.Count < 1) throw new Exception("adt_ops is empty at ExitOperations");
        }

        public void EnterOperation(ADTParser.OperationContext context) => op_sorts = new List<Sort>();
        public void ExitOperation(ADTParser.OperationContext context)
        {
            if (op_sorts.Count < 1) throw new Exception("op_sorts is empty at ExitOperation");
            string operationName = context.BEZ().GetText();
            Test_Operations_all_unique(operationName);
            Operation operation;
            if (context.Parent.GetChild(0).GetText() == "operations")
                operation = new Operation(operationName, op_sorts);
            else if (context.Parent.GetChild(0).GetText() == "constructors")
            {
                bool isBasic = true;
                for (int s = 0; s < op_sorts.Count - 1; s++)
                    if (op_sorts[s] == op_sorts.Last()) { isBasic = false; break; }
                if (isBasic) operation = new BasicConstructor(operationName, op_sorts);
                else operation = new BuildingConstructor(operationName, op_sorts);
                op_sorts.Last().SetConstructor((Constructor)operation);
            }
            else throw new Exception("ExitOperation with Context = " + context.GetText() + " in parent " + context.Parent.GetChild(0).GetText() + " was not able to find the sense of its life.");
            LH.adt_ops.Add(operation);
        }

        public void EnterVars([NotNull] ADTParser.VarsContext context) => LH.EnterVars();
        public void ExitVars([NotNull] ADTParser.VarsContext context) { }

        public void EnterVar([NotNull] ADTParser.VarContext context) { }
        public void ExitVar([NotNull] ADTParser.VarContext context)
            => LH.ExitVar(context.BEZ().GetText(), context.sort().BEZ().GetText());

        public void EnterAxioms([NotNull] ADTParser.AxiomsContext context) => LH.adt_axioms = new List<Axiom>();
        public void ExitAxioms([NotNull] ADTParser.AxiomsContext context)
        {
            if (LH.adt_axioms.Count < 1) throw new Exception("adt_axioms is empty at ExitAxioms");
        }

        public void EnterAxiom([NotNull] ADTParser.AxiomContext context) => LH.trees = new List<Tree>();
        public void ExitAxiom([NotNull] ADTParser.AxiomContext context)
        {
            string axiomName = context.BEZ().GetText();
            Test_Axioms_all_unique(axiomName);
            LH.ExitEq();
            LH.adt_axioms.Add(new Axiom(axiomName, LH.equation));
        }

        // Trees
        public void EnterMytree([NotNull] ADTParser.MytreeContext context) { }
        public void ExitMytree([NotNull] ADTParser.MytreeContext context) { }

        public void EnterEq_s([NotNull] ADTParser.Eq_sContext context) { }
        public void ExitEq_s([NotNull] ADTParser.Eq_sContext context) { }
        // classic Trees
        public void EnterClassicTree([NotNull] ADTParser.ClassicTreeContext context) { }
        public void ExitClassicTree([NotNull] ADTParser.ClassicTreeContext context) { }

        public void EnterTreeLeaf([NotNull] ADTParser.TreeLeafContext context) { }
        public void ExitTreeLeaf([NotNull] ADTParser.TreeLeafContext context)
            => LH.ExitTreeLeaf(context.BEZ().GetText(),new Listener_Helper.Coords(context.Start.Line,context.Start.Column));

        public void EnterTreeNode([NotNull] ADTParser.TreeNodeContext context) { }
        public void ExitTreeNode([NotNull] ADTParser.TreeNodeContext context)
            => LH.ExitTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(), new Listener_Helper.Coords(context.Start.Line, context.Start.Column));
        // function Trees
        public void EnterFTree([NotNull] ADTParser.FTreeContext context) { }

        public void ExitFTree([NotNull] ADTParser.FTreeContext context)
            => LH.ExitFTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(), new Listener_Helper.Coords(context.Start.Line, context.Start.Column));
        public void EnterFTreeBracketFree([NotNull] ADTParser.FTreeBracketFreeContext context) { }
        public void ExitFTreeBracketFree([NotNull] ADTParser.FTreeBracketFreeContext context)
            => LH.ExitFTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(), new Listener_Helper.Coords(context.Start.Line, context.Start.Column));

        public void EnterTreeBracket([NotNull] ADTParser.TreeBracketContext context) { }
        public void ExitTreeBracket([NotNull] ADTParser.TreeBracketContext context) { }

        public void EnterTreeBez([NotNull] ADTParser.TreeBezContext context) { }

        public void ExitTreeBez([NotNull] ADTParser.TreeBezContext context)
            => LH.ExitFTreeLeaf(context.BEZ().GetText(), new Listener_Helper.Coords(context.Start.Line, context.Start.Column));

        // useless Interface requirements
        public void EnterEveryRule(ParserRuleContext ctx) { }
        public void ExitEveryRule(ParserRuleContext ctx) { }
        public void VisitErrorNode(IErrorNode node) { }
        public void VisitTerminal(ITerminalNode node) { }

        /// <summary>
        /// Returns null if not able to find
        /// </summary>
        /// <param name="operationName"></param>
        /// <returns></returns>
        private Operation findOperationInADT_Ops(string operationName) => LH.findOperationInADT_Ops(operationName, 0, false);

        // Alle Offiziell notwendige Typtests

        /// <summary>
        /// This Test is activated always when a new Sort gets added to the sortList of the ADT,
        /// so as result all Sorts are unique.
        /// </summary>
        /// <param name="sortName"></param>
        private void Test_Sorts_all_unique(string sortName)
        {
            if (LH.findSortInADT_Sorts(sortName) != null) throw new Exception("The Sort " + sortName + " was defined more than 1 times in Sorts of ADT!");
        }
        /// <summary>
        /// This Test is activated always when a new Operation gets added to the operationList of the ADT,
        /// so as result all Operations are unique.
        /// </summary>
        /// <param name="operationName"></param>
        private void Test_Operations_all_unique(string operationName)
        {
            if (findOperationInADT_Ops(operationName) != null) throw new Exception("The Operation " + operationName + " was defined more than 1 times in Operations of ADT!");
        }
        /// <summary>
        /// This Test is activated always when a new Sort gets added to the sortList of an Operation,
        /// so as result every Sort of every Operation was defined in the ADT.
        /// </summary>
        /// <param name="sortName"></param>
        private void Test_Operation_all_Sorts_were_defined_in_ADT(string sortName)
        {
            if (LH.findSortInADT_Sorts(sortName) == null) throw new Exception("Found a Sort " + sortName + " in Operations that was not defined in Sorts!");
        }
        /// <summary>
        /// This Test is activated always when a new Axiom gets added to the axiomList of the ADT,
        /// so as result all Axioms are unique. 
        /// </summary>
        /// <param name="axiomName"></param>
        private void Test_Axioms_all_unique(string axiomName)
        {
            foreach (Axiom axiom in LH.adt_axioms)
            {
                if (axiom.name == axiomName) throw new Exception("The Axiom " + axiomName + " was defined more than 1 times in Axioms of ADT!");
            }
        }
        /// <summary>
        ///  This is done in LH.ExitEq() by Test_Equation_both_sides_have_same_Sort. 
        ///  This dummy Method is for knowing, that this Test is implemented.
        /// </summary>
        private void Test_Axiom_both_sides_have_same_Sort() {/* Nothing here need to be. */}
    }
}