﻿using Eingabeverwaltung.Metamodell;
using System;

namespace Ueberpruefung.CheckTransformation
{
    public class CheckUsedLemmaIsProved : ElementChecker
    {
        Lemma user;
        public CheckUsedLemmaIsProved(Lemma user) { this.user = user; }
        public override bool check(Element e)
        {
            if (e is TransformationStep ts)
            {
                if (ts.rule is Lemma l)
                {
                    bool result; string output = "";
                    if (l.IsProved is bool p) result = p; 
                    else throw new Exception("Tried to check if lemma " + l.name + " is proved while isProved is not set.");
                    if (result)
                    {
                        // This error must occure also when a main proof of a task has cyclic dependencies,
                        // but only if this usage is in that cycle
                        bool result2 = l.cyclicDependencies.Count == 0 || !l.cyclicDependencies.Contains(user);
                        if (!result2)
                        {
                            output = "{ ";
                            foreach (Lemma dep in user.cyclicDependencies) output += dep.name + " ";
                            output += "}";
                        }
                        result &= result2;
                    }
                    if (!result) ts.errors.Add(new Error(21,output));
                    return result;
                }
                else if (ts.rule is InductionHypothesis ih)
                {
                    bool result; 
                    if (ih.IsProved is bool p) result = p; 
                    else throw new Exception("Tried to check if inductionhypothesis is proved while isProved is not set.");
                    return result;
                }
                else return true;

            }else throw new InvalidOperationException();
        }
    }
}
